#ifdef _WIN32

#include <stdio.h>
#include "conv.h"

#define	BF_SIZ		2048
#define	bool		char
#define	true		1
#define	false		0

typedef unsigned char byte;

static
void	cConvSJ(),cConvJS(),cConvES(),cConvSE(),cConvAS();

struct _ConvRoutine Converters[] = {
	{'j', cConvJS, cConvSJ, "\033$@", "\033(J"},
	{'N', cConvJS, cConvSJ, "\033$B", "\033(J"},
	{'n', cConvJS, cConvSJ, "\033$B", "\033(B"},
	{'E', cConvES, cConvSE, NULL,     NULL    },
	{'S', NULL,    NULL,    NULL,     NULL    },
	{'J', cConvJS, cConvSJ, "\033$@", "\033(H"},
	{'U', cConvJS, cConvSJ, "\033$",  "\033+" },
	{'m', cConvJS, cConvSJ, "\033$@", "\033(B"},
	{'I', NULL,    NULL,    NULL,     NULL    },
	{'Z', cConvJS, cConvSJ, "\032p",  "\032q" },
	{'P', cConvJS, cConvSJ, "\033K",  "\033H" },
	{'A', cConvAS, cConvSJ, "\033$B", "\033(B"},
	{'-',NULL,    NULL,    NULL,     NULL    },
};

static byte	mbuf[BF_SIZ];

Converter DefaultConverter = Code_A;
Converter NoConv = Code_None;

Converter
getConverter(code)
byte code;
{
	int i;
	Converter cv = NULL;

	for (i = 0; &Converters[i] != NoConv; i++) {
		if (Converters[i].key == code) {
			cv = &Converters[i];
			break;
		}
	}
	return cv;
}

conv(is,os,fc,tc)
byte	*is,*os;
Converter fc,tc;
{
	if (fc == NULL || tc == NULL) {
		strcpy(os,is);
		return;
	}
	if (fc->from_routine && tc->to_routine) {
		(*fc->from_routine)(is,mbuf,fc->s_in,fc->s_out);
		(*tc->to_routine)(mbuf,os,tc->s_in,tc->s_out);
	}
	else if (fc->from_routine) {
		(*fc->from_routine)(is,os,fc->s_in,fc->s_out);
	}
	else if (tc->to_routine) {
		(*tc->to_routine)(is,os,tc->s_in,tc->s_out);
	}
	else
		strcpy(os,is);
}

static
putSFT(pptr,code)	/* Put Shift in/out code */
byte	**pptr,*code;
{
	while (*code != '\0') 
		*((*pptr)++) = *(code++);
}

static
byte getSLb(ptr,ub)	/* Get Shift-JIS Lower byte */
byte	*ptr,*ub;
{	byte	c = *ptr;

	*ub <<= 1;
	if (c < 0x9f) {
		if (c > 0x7e)
			c--;
		*ub -= 1;
		c -= 0x3f;
	}
	else {
		c -= 0x9e;
	}
	return c;
}

/* Convert Shift-JIS to JIS */
static
void cConvSJ(is,os,SIcode,SOcode)
byte	*is,*os;
byte *SIcode,*SOcode;
{	byte	c,km = 0,ub,lb,*p,*q = os;

	for (p = is; *p != '\0'; p++) {
		c = *p;
		if (0x80 <= c && c <= 0x9f) {
			if (km == 0) {
				putSFT(&q,SIcode);
				km = 1;
			}
			p++;
			ub = c-0x80;
			lb = getSLb(p,&ub);
			ub += 0x20;
			lb += 0x20;
			if (!SIcode) {    /* EUC */
				ub |= 0x80;
				lb |= 0x80;
			}
			*(q++) = ub;
			*(q++) = lb;
		}
		else if (0xe0 <= c && c <= 0xef) {
			if (km == 0) {
				putSFT(&q,SIcode);
				km = 1;
			}
			p++;
			ub = (c & 0x7f)-0x40;
			lb = getSLb(p,&ub);
			ub += 0x20;
			lb += 0x20;
			if (!SIcode) {    /* EUC */
				ub |= 0x80;
				lb |= 0x80;
			}
			*(q++) = ub;
			*(q++) = lb;
		}
		else {
			if (km == 1) {
				putSFT(&q,SOcode);
				km = 0;
			}
			*(q++) = c;
		}
	}
 	if (km == 1)
 		putSFT(&q,SOcode);
	*q = '\0';
}

/* Convert JIS to Shift-JIS */
static
void cConvJS(is,os,SIcode,SOcode)
byte *is,*os;
byte *SIcode,*SOcode;
{	byte	*ip = is,*op = os;
	byte	kmode = 0;
	int	il = strlen(SIcode),
		ol = strlen(SOcode);

	while (*ip != '\0') {
		if (strncmp(ip,SOcode,ol) == 0) {
			ip += ol;
			kmode = 0;
			continue;
		}
		if (strncmp(ip,SIcode,il) == 0) {
			ip += il;
			kmode = 1;
			continue;
		}
		if (kmode) {
			put_js(*ip,*(ip+1),op);
			ip += 2;
			op += 2;
		}
		else {
			*(op++) = *(ip++);
		}
	}
	*op = '\0';
}

/*
 * Convert Shift-JIS to EUC
 */
static
void cConvSE(is,os)
byte	*is,*os;
{	byte	c,km = 0,ub,lb,*p,*q = os;

	for (p = is; *p != '\0'; p++) {
		c = *p;
		if (0x80 <= c && c <= 0x9f) {
			if (km == 0) {
				km = 1;
			}
			p++;
			ub = c-0x80;
			lb = getSLb(p,&ub);
			ub += 0x20;
			lb += 0x20;
			ub |= 0x80;
			lb |= 0x80;
			*(q++) = ub;
			*(q++) = lb;
		}
		else if (0xe0 <= c && c <= 0xef) {
			if (km == 0) {
				km = 1;
			}
			p++;
			ub = (c & 0x7f)-0x40;
			lb = getSLb(p,&ub);
			ub += 0x20;
			lb += 0x20;
			ub |= 0x80;
			lb |= 0x80;
			*(q++) = ub;
			*(q++) = lb;
		}
		else {
			if (km == 1) {
				km = 0;
			}
			*(q++) = c;
		}
	}
	*q = '\0';
}

static
void cConvES(is,os)		/* Convert EUC to Shift-JIS */
byte	*is,*os;
{	byte	*ip = is,*op = os;
	byte	ub,lb;
	bool	firstByte = true;

	while (*ip != '\0') {
		ub = (byte)*(ip++);
		if (ub >= 0x80) {
			ub &= 0x7f;
			lb = (byte)*(ip++) & 0x7f;
			put_js(ub,lb,op);
			op += 2;
		}
		else {
			*(op++) = (byte)ub;
		}
	}
	*op = '\0';
}

/* Convert any JIS to Shift-JIS */
static
void cConvAS(is,os)
byte *is,*os;
{	byte	*ip = is,*op = os;
	byte	kmode = 0;

	while (*ip != '\0') {
		if (*ip == '\033') {
			ip++;
			switch(*ip) {
			case '(':
				ip++;
				switch(*ip) {
				case 'J':
				case 'B':
				case 'H':
				case '@':
					kmode = 0;
					ip++;
					break;
				default:
					*(op++) = '\033';
					*(op++) = '(';
				}
				break;
			case '$':
				ip++;
				switch(*ip) {
				case 'B':
				case '@':
					kmode = 1;
					ip++;
					break;
				default:
					*(op++) = '\033';
					*(op++) = '$';
				}
				break;
			default:
				*(op++) = '\033';
			}
		}
		if (kmode) {
			put_js(*ip,*(ip+1),op);
			ip += 2;
			op += 2;
		}
		else {
			*(op++) = *(ip++);
		}
	}
	*op = '\0';
}

put_js(ub,lb,op)
byte	ub,lb;
byte	*op;
{	
	ub -= 0x20;
	lb -= 0x20;
	if ((ub&1) == 0)
		lb += 0x5e;
	ub = ((ub-1)>>1)+0x81;
	lb += 0x3f;
	if (ub > 0x9f)
		ub += 0x40;
	if (lb > 0x7e)
		lb++;

	*op = ub;
	*(op+1) = lb;
}

static Converter
selectConverter(si,so)
char si,so;
{
	switch (si) {
	case '@':
		switch (so) {
		case 'H':
			return Code_J;
		case 'J':
			return Code_j;
		case 'B':
			return Code_m;
		default:
			return NULL;
		}
	case 'B':
		switch (so) {
		case 'J':
			return Code_N;
		case 'B':
			return Code_n;
		default:
			return NULL;
		}
	default:
		return NULL;
	}
}

Converter
checkShiftCode(buf,lastbyte)
byte	*buf;
int *lastbyte;
{	byte	code,*p,
		si = '\0',
		so = '\0';
	int nbyte = 1;
	int consistent = 1;

	for (p = buf; *p != '\0'; p++) {
		if (*p == '\033') {
			switch (*(++p)) {
			case '$':	/* to non-US-ASCII */
				p++;
				if (si != '\0' && si != *p)
					consistent = 0;
				si = *p;
				nbyte = 2;
				break;
			case '(':	/* to US-ASCII or X0201 ) */
				p++;
				if (so != '\0' && so != *p)
					consistent = 0;
				so = *p;
				nbyte = 1;
				break;
			}
		}
	}
	if (lastbyte)
		*lastbyte = nbyte;
	if (consistent)
		return selectConverter(si,so);
	else
		return Code_A;
}

Converter
guessMailCode(f,firstpos,curpos,lastpos)
FILE *f;
int firstpos,curpos;
{
	unsigned char c;
	char si = '\0', so = '\0';
	char xsi = '\0', xso = '\0';
	char bit8 = 0, cbit8 = 0;
	Converter guess = NULL;
	
	fseek(f,firstpos,0);
        while (1) {
		c = (unsigned char)getc(f);
		if (feof(f)) break;
		if (lastpos >= 0 && lastpos <= ftell(f))
			break;
		if (c & 0x80) {
			cbit8 = 1;
			if (bit8) {
				bit8 = 0;
				if (c < 0xa1) {
					guess = Code_S;
					break;
				}
			}
			else {
				bit8 = 1;
				if (c < 0xa1) {
					guess = Code_S;
					break;
				}
			}
		}
		else if (bit8) {
			guess = Code_S;
			break;
		}
		if (c == '\033') {
			c = getc(f);
			if (c == '$') {
				si = getc(f);
				if (si != '@' && si != 'B')
					si = '\0';
				else if (xsi != '\0' && xsi != si) {
					guess = Code_A;
					break;
				}
				xsi = si;
			}
			else if (c == '(') {
				so = getc(f);
				if (so != 'J' && so != 'B' && so != 'H')
					so = '\0';
				else if (xso != '\0' && xso != so) {
					guess = Code_A;
					break;
				}
				xso = so;
			}
		}
	}
	fseek(f,curpos,0);
	if (guess)
		return guess;
	else if (cbit8)
		return Code_E;
	guess = selectConverter(si,so);
	if (guess == NULL)
                guess = NoConv;
        return guess;
}

#endif
