/*
 * MIME support by Akinori ITO
 */
#include <stdio.h>

#define MIME_ENCODED_LINE_LIMIT	80
#define MIME_ENCODED_WORD_LENGTH_OFFSET 18
#define MIME_ENCODED_WORD_LENGTH_ESTIMATION(x) \
	(((x)+2)*4/3+MIME_ENCODED_WORD_LENGTH_OFFSET)
#define MIME_DECODED_WORD_LENGTH_ESTIMATION(x) \
	(((x)-MIME_ENCODED_WORD_LENGTH_OFFSET)/4*3)
#define J_CHARSET "ISO-2022-JP"

#define BAD_BASE64 255

#define LINELEN 4096

char *malloc();

#define strequal(a,b) (strcasecmp((a),(b))==0)

static
unsigned char
c2e(x)
char x;
{
	if ('A'<= x && x <='Z') return (x)-'A';
	if ('a'<= x && x <='z') return (x)-'a'+26;
	if ('0'<= x && x <='9') return (x)-'0'+52;
	if (x =='+') return 62;
	if (x == '/') return 63;
	return BAD_BASE64;
}

static
int
ha2d(x,y)
char x,y;
{
	int r = 0;

	if      ('0' <= x && x <= '9') r = x-'0';
	else if ('A' <= x && x <= 'F') r = x-'A'+10;
	else if ('a' <= x && x <= 'f') r = x-'a'+10;

	r <<= 4;

	if      ('0' <= y && y <= '9') r += y-'0';
	else if ('A' <= y && y <= 'F') r += y-'A'+10;
	else if ('a' <= y && y <= 'f') r += y-'a'+10;

	return r;

}

char *
decodeB(w,a)
char *w,*a;
{
	unsigned char c[4];
	char *wp = w, *ap = a;
	char d[3];
	int i,n_pad;

	n_pad = 0;
	while (1) {
		for (i = 0; i < 4; i++) {
			c[i] = *(wp++);
			if (*wp == '\0' || *wp == '?') {
				i++;
				for (; i < 4; i++) {
					c[i] = '=';
				}
				break;
			}
		}
		if (c[3] == '=') {
			n_pad++;
			c[3] = 'A';
			if (c[2] == '=') {
				n_pad++;
				c[2] = 'A';
			}
		}
		for (i = 0; i < 4; i++) {
			c[i] = c2e(c[i]);
			if (c[i] == BAD_BASE64) {
				return w;
			}
		}
		d[0] = ((c[0] << 2) | (c[1] >> 4));
		d[1] = ((c[1] << 4) | (c[2] >> 2));
		d[2] = ((c[2] << 6) | c[3]);
		for (i = 0; i < 3-n_pad; i++) {
			*ap = d[i];
			ap++;
		}
		if (n_pad || *wp == '\0' || *wp == '?')
			break;
	}
	*ap = '\0';
	return wp;
}

int
decodeB2file(w,fn)
char *w,*fn;
{
	unsigned char c[4];
	char *wp = w;
	char d[3];
	int i,n_pad;
        FILE *f;

	f = fopen(fn,"w");
	if (f == NULL) return -1;
	n_pad = 0;
	while (1) {
		for (i = 0; i < 4; i++) {
			c[i] = *(wp++);
			if (*wp == '\0' || *wp == '?') {
				i++;
				for (; i < 4; i++) {
					c[i] = '=';
				}
				break;
			}
		}
		if (c[3] == '=') {
			n_pad++;
			c[3] = 'A';
			if (c[2] == '=') {
				n_pad++;
				c[2] = 'A';
			}
		}
		for (i = 0; i < 4; i++) {
			c[i] = c2e(c[i]);
			if (c[i] == BAD_BASE64) {
				return -1;
			}
		}
		d[0] = ((c[0] << 2) | (c[1] >> 4));
		d[1] = ((c[1] << 4) | (c[2] >> 2));
		d[2] = ((c[2] << 6) | c[3]);
		for (i = 0; i < 3-n_pad; i++) {
			fputc(d[i],f);
		}
		if (n_pad || *wp == '\0' || *wp == '?')
			break;
	}
        fclose(f);
	return 0;
}

char *
decodeQ(w,a)
char *w,*a;
{
	for (; *w != '\0' && *w != '?'; w++) {
		if (*w == '=') {
			w++;
			*(a++) = ha2d(*w,*(w+1));
			w++;
		}
		else if (*w == '_') {
			*(a++) = ' ';
		}
		else
			*(a++) = *w;
	}
	*a = '\0';
	return w;
}

int
decodeQ2file(w,fn)
char *w,*fn;
{
        FILE *f;
        f = fopen(fn,"w");
        if (f == NULL) return -1;
	for (; *w != '\0' && *w != '?'; w++) {
		if (*w == '=') {
			w++;
			fputc(ha2d(*w,*(w+1)),f);
			w++;
		}
		else if (*w == '_') {
			fputc(' ',f);
		}
		else
			fputc(*w,f);
	}
	return 0;
}

#define CHARSET_LEN 32

char*
decodeWord(ow,a)
char *ow,*a;
{
	char charset[CHARSET_LEN];
	char *p;
	char *w = ow;
	char method;
	
	if (*w != '=' || *(w+1) != '?') 
		goto convert_fail;
	w += 2;
	for (p = charset; *w != '?' && p < charset+CHARSET_LEN; w++) {
		if (*w == '\0')
			goto convert_fail;
		*(p++) = *w;
	}
	if (p == charset+CHARSET_LEN)
		goto convert_fail;
	*p = '\0';
	if (!strequal(charset,J_CHARSET)) {
		/* NOT ISO-2022-JP encoding ... don't convert */
		goto convert_fail;
	}
	w++;
	method = *(w++);
	if (*w != '?') goto convert_fail;
	w++;
	switch(method) {
	    case 'B':
		p = decodeB(w,a);
		break;
	    case 'Q':
		p = decodeQ(w,a);
		break;
	    default:
		goto convert_fail;
	}
	if (p == w) goto convert_fail;
	w = p;
	if (*w == '?') {
		w++;
		if (*w == '=')
			w++;
	}
	return w;

convert_fail:
	*a = '\0';
	return ow;
}

static char Base64Table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

char*
encodeB(a,w)
char *a,*w;
{
	char *ww = w;
	unsigned char d[3];
	unsigned char c1,c2,c3,c4;
	int i,n_pad;

	while (1) {
		if (*a == '\0')
			break;
		n_pad = 0;
		d[1] = d[2] = 0;
		for (i = 0; i < 3; i++) {
			d[i] = a[i];
			if (a[i] == '\0') {
				n_pad = 3-i;
				break;
			}
		}
		c1 = d[0] >> 2;
		c2 = (((d[0] << 4) | (d[1] >> 4)) & 0x3f);
		if (n_pad == 2) {
			c3 = c4 = 64;
		}
		else if (n_pad == 1) {
			c3 = ((d[1] << 2) & 0x3f);
			c4 = 64;
		}
		else {
			c3 = (((d[1] << 2) | (d[2] >> 6)) & 0x3f);
			c4 = (d[2] & 0x3f);
		}
		*(w++) = Base64Table[c1];
		*(w++) = Base64Table[c2];
		*(w++) = Base64Table[c3];
		*(w++) = Base64Table[c4];
		if (n_pad)
			break;
		a += 3;
	}
	*w = '\0';
	return w;
}

char *
encodeWord(a,w,charset)
char *a,*w,*charset;
{
	char *p;

	*(w++) = '=';
	*(w++) = '?';
	for (p = charset; *p; p++)
		*(w++) = *p;
	*(w++) = '?';
	*(w++) = 'B';
	*(w++) = '?';
	w = encodeB(a,w);
	*(w++) = '?';
	*(w++) = '=';
	*w = '\0';
	return w;
}

/*
 * convert MIME encoded string to the original one
 */
decodeMIME(orgstr,cnvstr)
char *orgstr,*cnvstr;
{
	char *org = orgstr, *cnv = cnvstr;
	char *org0,*p;
	while (1) {
		if (*org == '\0')
			break;
		if (*org == '=' && *(org+1) == '?') {
			p = decodeWord(org,cnv);
			if (org == p) {  /* Convert failure */
				strcpy(cnvstr,orgstr);
				return;
			}
			org = p;
			while (*cnv) cnv++;
			for (org0 = org; *org0 == ' ' || *org0 == '\t'; org0++);
			if (org0[0] == '=' && org0[1] == '?')
				org = org0;
		}
		else 
			*(cnv++) = *(org++);
	}
	*cnv = '\0';
}

encodeMIME(org,cnv)
char *org,*cnv;
{
	char *o,*p,*c,*c0;
	char *buf = malloc(LINELEN);
	int i,len,d_est;

	c = c0 = cnv;
	o = org;
	while (1) {
		if (o[0] == '\033' && o[1] == '$' && o[2] != '\0') {
			if (o > org && o[-1] != ' ' && o[-1] != '\t')
				*c++ = ' ';
			/* MIME encode start */
			p = o+3;
			while (*p != '\0' && *p != '\n' &&
			       !(p[0] == '\033' && p[1] == '(' && p[2] != '\0')) {
				p++;
			}
			if (*p == '\033')
				p += 3;
			len = c-c0;
			while (len+MIME_ENCODED_WORD_LENGTH_ESTIMATION(p-o) >
			    MIME_ENCODED_LINE_LIMIT) {
				d_est = MIME_DECODED_WORD_LENGTH_ESTIMATION(MIME_ENCODED_LINE_LIMIT-len);
				if ((d_est & 1) == 1) d_est++;
				if (d_est > 0) {
					for (i = 0; i < d_est; i++) {
						if (*o == '\033' && i > d_est-3) {
							/* Avoid folding within the escape sequennce */
							break;
						}
						buf[i] = *(o++);
					}
					buf[i] = '\0';
					c = encodeWord(buf,c,J_CHARSET);
				}
			        *c++ = '\n';
				c0 = c;
			        *c++ = '\t';
				len = 1;
			}
			for (i = 0; o < p; o++)
				buf[i++] = *o;
			if (i > 0) {
				buf[i] = '\0';
				c = encodeWord(buf,c,J_CHARSET);

				if (o[1] != ' ' && o[1] != '\t')
					*c++ = ' ';
			}

		}
		else if (*o == '\0')
			break;
		else
			*c++ = *o++;
	}
	*c = '\0';
	free(buf);
}
